from time import sleep

from pbx_gs_python_utils.utils.Files import Files
from pbx_gs_python_utils.utils.Misc import Misc

from osbot_aws.Globals import Globals
from osbot_aws.apis.Lambda import Lambda
from osbot_aws.apis.S3 import S3
from osbot_aws.apis.test_helpers.Temp_Aws_Roles import Temp_Aws_Roles
from osbot_aws.helpers.IAM_Role import IAM_Role


class Temp_Lambda:
    def __init__(self):
        self.name           = "temp_lambda_{0}".format(Misc.random_string_and_numbers())
        self.temp_lambda    = Lambda(self.name)
        self.tmp_folder     = Temp_Folder_Code(self.name)
        self.role_arn       = Temp_Aws_Roles().for_lambda_invocation()
        self.create_log     = None
        self.delete_on_exit = True
        self.lambda_name    = 'tmp_lambda_dev_test'
        self.s3_bucket      = Globals.lambda_s3_bucket
        self.s3_key         =  'unit_tests/lambdas/{0}.zip'.format(self.lambda_name)


    def __enter__(self):
        (
            self.temp_lambda.set_role       (self.role_arn)
                            .set_s3_bucket  (self.s3_bucket          )
                            .set_s3_key     (self.s3_key             )
                            .set_folder_code(self.tmp_folder.folder )
        )
        self.temp_lambda.upload()
        self.create_log = self.temp_lambda.create()
        assert self.temp_lambda.exists() == True
        return self

    def __exit__(self, type, value, traceback):
        if self.delete_on_exit:
            assert self.temp_lambda.delete() is True

    def invoke(self, params=None):
        return self.temp_lambda.invoke(params)

    def invoke_raw(self, params=None):
        return self.temp_lambda.invoke_raw(params)

    def set_s3_bucket(self, value):
        self.s3_bucket = value

class Temp_Folder_Code:
    def __init__(self, file_name):
        self.file_name    = file_name
        self.s3           = S3()
        self.folder       = Files.temp_folder('tmp_lambda_')
        self.lambda_code  = "def run(event, context): return 'hello {0}'.format(event.get('name'))"
        self.tmp_file     = None
        self.s3_bucket    = Globals.lambda_s3_bucket
        self.lambda_name  = 'tmp_lambda_dev_test'
        self.s3_key       = 'unit_tests/lambdas/{0}.zip'.format(self.lambda_name)
        self.create_temp_file()

    def create_temp_file(self, new_code=None):
        if new_code: self.lambda_code = new_code
        self.tmp_file = Files.path_combine(self.folder, '{0}.py'.format(self.file_name))
        Files.write(self.tmp_file, self.lambda_code)
        assert Files.exists(self.tmp_file)
        return self

    def s3_file_exists(self):
        return self.s3.file_exists(self.s3_bucket, self.s3_key)

    def delete_s3_file(self):
        self.s3.file_delete(self.s3_bucket, self.s3_key)
        assert self.s3_file_exists() is False
        return self



import calendar
import datetime
import time

import boto3
import botocore
from pbx_gs_python_utils.utils.Dev import Dev

from osbot_aws.apis.Boto_Helpers import Boto_Helpers
from osbot_aws.apis.Session import Session


class Logs(Boto_Helpers):
    def __init__(self, group_name, stream_name=None):
        self.log_group_name = group_name
        self.stream_name    = stream_name
        self._logs          = None

    # helper methods
    def logs(self):
        if self._logs is None:
            self._logs = Session().client('logs')
        return self._logs

    def timestamp_utc_now(self):        # bug: this is not working 100% (I'm getting 1h early)
        #calendar.timegm(time.gmtime())
        return int(datetime.datetime.utcnow().strftime('%s')) * 1000

    # main methods

    def create(self):
        self.group_create()
        self.stream_create()
        return self

    def event_add(self, message, timestamp=None, sequence_token=None):
        if timestamp is None: timestamp = self.timestamp_utc_now()
        return self.events_add([{'message': message, 'timestamp':timestamp}], sequence_token)

    def events_add(self, events, sequence_token=None):
        try:
            if sequence_token:
                result  = self.logs().put_log_events(logGroupName=self.log_group_name,logStreamName=self.stream_name,logEvents=events, sequenceToken=sequence_token)
            else:
                result  = self.logs().put_log_events(logGroupName=self.log_group_name,logStreamName=self.stream_name,logEvents=events)
            nextSequenceToken = result.get('nextSequenceToken')
            if result.get('rejectedLogEventsInfo') is not None:
                return {'status': 'error', 'nextSequenceToken' : nextSequenceToken, 'data': result.get('rejectedLogEventsInfo')}
            return {'status': 'ok', 'nextSequenceToken' : nextSequenceToken}
        except Exception as error:
            return {'status': 'error', 'data': '{0}'.format(error)}


    def events(self):
        return self.logs().get_log_events(logGroupName=self.log_group_name, logStreamName=self.stream_name)

    def group_create(self):
        if self.group_exists(): return False
        self.logs().create_log_group(logGroupName=self.log_group_name)
        return self.group_exists()

    def group_delete(self):
        if self.group_exists() is False: return False
        self.logs().delete_log_group(logGroupName=self.log_group_name)
        return self.group_exists() is False

    def group_exists(self):
        return self.group_info() is not None

    def group_info(self):
        for log_group in self.logs().describe_log_groups(logGroupNamePrefix= self.log_group_name).get('logGroups'):
            if log_group.get('logGroupName') == self.log_group_name:
                return log_group
        return None

    def groups(self):
        groups = {}
        for group in  self.invoke_using_paginator(self.logs(), 'describe_log_groups', 'logGroups'):
            groups[group.get('logGroupName')] = group
        return groups

    def stream_create(self):
        if self.stream_exists() is True: return False
        self.logs().create_log_stream(logGroupName=self.log_group_name, logStreamName= self.stream_name)
        return self.stream_exists()

    def stream_delete(self):
        if self.stream_exists() is False: return False
        self.logs().delete_log_stream(logGroupName=self.log_group_name, logStreamName= self.stream_name)
        return self.stream_exists() is False

    def stream_exists(self):
        return len(self.streams()) > 0

    def streams(self):
        if self.stream_name is None:
            return self.logs().describe_log_streams(logGroupName=self.log_group_name).get('logStreams')
        return self.logs().describe_log_streams(logGroupName=self.log_group_name, logStreamNamePrefix=self.stream_name).get('logStreams')

    def messages(self):
        messages = []
        for event in self.logs().get_log_events(logGroupName=self.log_group_name, logStreamName=self.stream_name).get('events'):
            messages.append(event.get('message'))
        return messages


import gzip
import os
import boto3
import tempfile
from   io import BytesIO
from   gzip import GzipFile
from   botocore.errorfactory import ClientError
from   urllib.parse          import urlparse

from    pbx_gs_python_utils.utils.Files import Files

from osbot_aws.apis.Session import Session


class S3:
    def __init__(self):
        self.boto_client_s3  = None
        self.boto_resource_s3 = None
        self.tmp_file_folder = 's3_temp_files'

    # helpers
    def s3(self):
        if self.boto_client_s3 is None : self.boto_client_s3 = Session().client('s3')
        return self.boto_client_s3

    def s3_resource(self):
        if self.boto_resource_s3 is None : self.boto_resource_s3 = Session().resource('s3')
        return self.boto_resource_s3

    def s3_bucket_notification(self,bucket_name):
        return self.s3_resource().BucketNotification(bucket_name)

    # main methods
    def bucket_arn(self,bucket):
        return 'arn:aws:s3:::{0}'.format(bucket)

    def bucket_create(self, bucket,region):
        try:
            location=self.s3().create_bucket(Bucket=bucket,CreateBucketConfiguration={'LocationConstraint': region }).get('Location')
            return { 'status':'ok', 'data':location}
        except Exception as error:
            return {'status': 'error', 'data': '{0}'.format(error)}

    def bucket_delete(self, bucket):
        if self.bucket_exists(bucket) is False:
            return False
        self.s3().delete_bucket(Bucket=bucket)
        return self.bucket_exists(bucket) is False

    def bucket_exists(self, bucket):
        try:
            self.s3().head_bucket(Bucket=bucket)
            return True
        except:
            return False



    def buckets(self):
        data = []
        for bucket in self.s3().list_buckets().get('Buckets'):
            data.append(bucket['Name'])
        return sorted(data)
        #return sorted(list({ bucket['Name'] for bucket in self.s3().list_buckets().get('Buckets')}))
        #

    def _files_raw              (self, bucket, prefix='', filter=''                 ):
        params = {'Bucket': bucket}
        if isinstance(prefix, str):                                            # if prefix is a string, filter on list_objects_v2
            params['Prefix'] = prefix

        while True:
            resp = self.s3().list_objects_v2(**params)
            try:
                contents = resp['Contents']
            except KeyError:
                return

            for obj in contents:
                key = obj['Key']
                if (not filter) or (filter and filter in key):
                    yield obj
            try:
                params['ContinuationToken'] = resp['NextContinuationToken']     # token for next batch of 1000 results
            except KeyError:
                break

    def find_files              (self, bucket, prefix='', filter=''                 ):
        return list({ item['Key'] for item in self._files_raw(bucket, prefix, filter) })

        #

    def file_contents           (self, bucket, key                                  ):
        obj = self.s3().get_object(Bucket=bucket, Key=key)                    # get object data from s3
        return obj['Body'].read().decode('utf-8')                               # extract body and decode it

    def file_contents_as_gzip   (self, bucket, key                                  ):
        obj = self.s3().get_object(Bucket=bucket, Key=key)                    # get object data from s3
        bytestream = BytesIO(obj['Body'].read())
        return GzipFile(None, 'rb', fileobj=bytestream).read().decode('utf-8')

    def file_download           (self, bucket, key , use_cache = False               ):
        if bucket and key:
            tmp_file = '/tmp/' + key.replace('/','_')
            if self.file_download_to(bucket, key, tmp_file, use_cache):
                return tmp_file
        return None

    def file_download_and_delete(self, bucket, key):
        tmp_file = '/tmp/' + key.replace('/','_')
        if self.file_download_to(bucket, key, tmp_file):
            self.file_delete(bucket,key)
            return tmp_file
        return None

    def file_download_to        (self, bucket, key, target_file, use_cache = False  ):
        try:
            if use_cache is True:
                if os.path.exists(target_file):
                    return True
            else:
                if os.path.exists(target_file):
                    os.remove(target_file)
            self.s3().download_file(bucket, key, target_file)
            return os.path.exists(target_file)
        except ClientError:                                     # when file not found
            return False

    def file_copy               (self, src_bucket, src_key, dest_bucket, dest_key   ):
        return self.s3().copy({'Bucket': src_bucket, 'Key': src_key}, dest_bucket,dest_key )

    def file_create_from_string_as_gzip (self, file_contents, bucket, key                   ):
        tmp_path = tempfile.NamedTemporaryFile().name

        with gzip.open(tmp_path,'w') as file:                       # create gzip on tmp_file
            file.write(file_contents.encode())                      # write contents
        self.s3().upload_file(tmp_path, bucket,key)               # upload file using s3 api
        os.remove(tmp_path)                                         # delete tmp file
        return True

    def file_create_from_string (self, file_contents, bucket, key                   ):
        with tempfile.NamedTemporaryFile() as temp:                     # use tempfile api to create temp file
            temp.write(str.encode(file_contents))                       # write contents
            temp.flush()                                                # flush contents to make sure everything has been written
            self.s3().upload_file(temp.name, bucket,key)              # upload file using s3 api
            return True

    def file_delete             (self, bucket, key                                  ):
        result = self.s3().delete_object(Bucket=bucket, Key=key)              # delete file from s3
        return result.get('ResponseMetadata').get('HTTPStatusCode') == 204

    def file_details(self, bucket, key):
        return self.s3().head_object(Bucket = bucket, Key= key);

    def file_size_in_Mb(self, bucket, key):
        details = self.file_details(bucket,key)
        return '{0:.1f} Mb'.format(int(details['ResponseMetadata']['HTTPHeaders']['content-length']) / 1024 / 1024)

    def file_exists(self, bucket, key):
        try:
            self.s3().head_object(Bucket=bucket, Key=key)                     # calls head_object function
            return True                                                       # if it works, file exists (return true)
        except:                                                               # if we get an exception
            return False                                                      # file doesn't exists (return false)

    def file_move               (self,src_bucket, src_key, dest_bucket, dest_key):
        self.file_copy(src_bucket, src_key, dest_bucket, dest_key)
        self.file_delete(src_bucket, src_key)
        return self.file_exists(dest_bucket, dest_key)

    def file_upload             (self,file , bucket, folder):
        if not os.path.isfile(file):                                            # check that file to upload exists locally
            return None                                                         # if not return None

        key = os.path.join(folder, os.path.basename(file))                      # create key from folder and file's filename

        self.file_upload_to_key(file, bucket, key)                                # upload file
        return key                                                              # return path to file uploaded (if succeeded)

    def file_upload_to_key(self, file, bucket, key):
        self.s3().upload_file(file, bucket, key)                                # upload file
        return True                                                             # return true (if succeeded)

    def file_upload_as_temp_file(self, file, bucket):
        key = '{0}/{1}'.format(self.tmp_file_folder, Files.temp_filename(Files.file_extension(file)))
        self.file_upload_to_key(file, bucket, key)
        return key


    def folder_upload (self, folder, s3_bucket, s3_key):
        file = Files.zip_folder(folder)
        self.file_upload_to_key(file, s3_bucket, s3_key)
        os.remove(file)
        return self



# def split_s3_url(s3_url):
#     url = urlparse(s3_url)
#     bucket = url.netloc
#     path = url.path.lstrip('/')
#     return bucket, path


from time import sleep

import boto3

from pbx_gs_python_utils.utils.Dev import Dev

from osbot_aws.apis.Cloud_Watch import Cloud_Watch
from osbot_aws.apis.IAM import IAM
from osbot_aws.apis.Session import Session


class Fargate:

    def __init__(self,account_id):
        self.ecs        = Session().client('ecs')
        self.account_id = account_id

    def cluster_create(self, cluster_name):
        return self.ecs.create_cluster(clusterName=cluster_name)
        
    def cluster_delete(self, cluster_arn):
        return self.ecs.delete_cluster(cluster=cluster_arn)

    def clusters(self):
        return self.ecs.list_clusters().get('clusterArns')

    def task_get_log_details(self, task_name):
        log_group_name = "awslogs-{0}".format(task_name)
        log_group_region = "eu-west-2"
        log_group_stream_prefix = "awslogs-example"
        cloud_watch = Cloud_Watch()
        if cloud_watch.log_group_exists(log_group_name) is False:
            cloud_watch.log_group_create(log_group_name)
        return log_group_name,log_group_region,log_group_stream_prefix

    def task_create(self, task_name,task_role_arn,execution_role_arn):
        (log_group_name, log_group_region, log_group_stream_prefix) = self.task_get_log_details(task_name)

        kwargs = {
            'family'              : task_name,
            'taskRoleArn'         : task_role_arn,
            'executionRoleArn'    : execution_role_arn,
            'cpu'                 : "1024",
            'memory'              : "2048",
            'networkMode'         : 'awsvpc',
            "containerDefinitions": [ {
                'name'            : 'gs-docker-codebuild',
                'image'           : '{0}.dkr.ecr.eu-west-2.amazonaws.com/gs-docker-codebuild:latest'.format(self.account_id),
                "logConfiguration": {
                                        "logDriver": "awslogs",
                                        "options": {
                                                    "awslogs-group"        : log_group_name,
                                                    "awslogs-region"       : log_group_region,
                                                    "awslogs-stream-prefix": log_group_stream_prefix
                                                   }},

            }],
            "requiresCompatibilities": [ 'FARGATE'] }

        try:
            return self.ecs.register_task_definition(**kwargs).get('taskDefinition')
        except Exception as error:
            return "{0}".format(error)

    def task_run(self, cluster, task_arn,subnet_id,security_group):
        kwargs = {
                    'cluster'             : cluster ,
                    'taskDefinition'      : task_arn,
                    'launchType'          : 'FARGATE',
                    'networkConfiguration': { 'awsvpcConfiguration': { 'subnets'        : [ subnet_id],
                                                                       'securityGroups' : [ security_group],
                                                                       'assignPublicIp' : 'ENABLED'  } },
                 }

        try:
            return self.ecs.run_task(**kwargs).get('tasks')[0]
        except Exception as error:
            return "{0}".format(error)

    def task_delete(self, task_arn):
        return self.ecs.deregister_task_definition(taskDefinition=task_arn)

    def task_details(self, cluster, task_arn):
        result = self.ecs.describe_tasks(cluster=cluster, tasks=[task_arn])
        return result.get('tasks').pop()

    def task_wait_for_completion(self, cluster, task_arn, sleep_for=1, max_attempts=30, log_status=False):
        for i in range(0,max_attempts):
            build_info    = self.task_details(cluster, task_arn)
            last_Status  = build_info.get('lastStatus')
            #current_phase = build_info.get('currentPhase')
            if log_status:
                Dev.pprint("[{0}] {1}".format(i,last_Status))
            if last_Status == 'DEPROVISIONING' or last_Status == 'STOPPED':
                return build_info
            sleep(sleep_for)
        return build_info

    def tasks(self):
        return self.ecs.list_task_definitions().get('taskDefinitionArns')



    def policy_create_for_task_role(self, role_name):
        policy = { "Version"    : "2008-10-17",
                   "Statement"  : [ { "Effect": "Allow",
                                      "Principal": { "Service": "ecs-tasks.amazonaws.com"},
                                      "Action": "sts:AssumeRole" }]}
        IAM(role_name=role_name).role_create(policy)

    def policy_create_for_execution_role(self, role_name):
        region = 'eu-west-2'
        cloud_watch_arn = "arn:aws:logs:{0}:{1}:log-group:awslogs-*".format(region, self.account_id)
        role_policy = { "Version"   : "2008-10-17",
                        "Statement" : [ { "Effect": "Allow",
                                          "Principal": { "Service": "ecs-tasks.amazonaws.com"},
                                          "Action": "sts:AssumeRole" }]}
        policy      = { "Version"  : "2012-10-17",
                        "Statement": [{   "Effect"  :  "Allow"                              ,
                                          "Action"  : [  "ecr:GetAuthorizationToken"        ,
                                                         "ecr:BatchCheckLayerAvailability"  ,
                                                         "ecr:GetDownloadUrlForLayer"       ,
                                                         "ecr:GetRepositoryPolicy"          ,
                                                         "ecr:DescribeRepositories"         ,
                                                         "ecr:ListImages"                   ,
                                                         "ecr:DescribeImages"               ,
                                                         "ecr:BatchGetImage"               ],
                                            "Resource": "*"     },
                                      {
                                           "Effect"  : "Allow",
                                           "Action"  : [ "logs:CreateLogStream" ,
                                                         "logs:PutLogEvents"   ],
                                           "Resource": [ cloud_watch_arn ]}]}

        policy_name = 'policy_for_{0}'.format(role_name)
        iam = IAM(role_name=role_name)

        iam.role_create(role_policy)
        iam.policy_delete(policy_name)
        policy_arn = iam.policy_create(policy_name,policy).get('policy_arn')

        iam.role_policy_attach(policy_arn)






import datetime

from pbx_gs_python_utils.utils.Assert import Assert


class Temp_Assert:

    def __init__(self,target):
        self._assert = Assert(target)

    def is_today(self):
        assert type(self._assert.target) == datetime.datetime
        assert str(self._assert.target.date())  == str(datetime.datetime.utcnow().date())


import boto3
from   boto3    import resource
from pbx_gs_python_utils.utils.Misc import Misc

from osbot_aws.apis.Session import Session


class Dynamo:
    def __init__(self):
        self.resource   = resource('dynamodb')
        self._dynamo    = None
        self._streams   = None

    # helpers

    def dynamo(self):
        if self._dynamo is None:
            self._dynamo = Session().client('dynamodb')
        return self._dynamo

    def dynamo_streams(self):
        if self._streams is None:
            self._streams = Session().client('dynamodbstreams')
        return self._streams
    # main methods

    def create(self, table_name, key, with_streams=False):
        keySchema             = [ {'AttributeName'    : key        , 'KeyType'           : 'HASH' } ]
        attributeDefinitions  = [ {'AttributeName'    : key        , 'AttributeType'     : 'S'    } ]
        provisionedThroughput = { 'ReadCapacityUnits' : 5          , 'WriteCapacityUnits': 5      }
        kwargs   = { 'TableName'            : table_name           ,
                     'KeySchema'            : keySchema            ,
                     'AttributeDefinitions' : attributeDefinitions ,
                     'ProvisionedThroughput': provisionedThroughput }
        if with_streams:
            kwargs['StreamSpecification'] = {'StreamEnabled': True, 'StreamViewType': 'NEW_IMAGE' }
        self.dynamo().create_table( **kwargs)

        self.dynamo().get_waiter('table_exists') \
            .wait(TableName=table_name, WaiterConfig={'Delay': 5, 'MaxAttempts': 10})
        return self

    def delete(self, table_name):
        self.dynamo().delete_table(TableName = table_name)
        self.dynamo().get_waiter('table_not_exists')      \
                   .wait(TableName=table_name, WaiterConfig={'Delay': 10, 'MaxAttempts':10 })
        return self

    def list(self):
        return self.dynamo().list_tables()['TableNames']

    def streams(self):
        return self.dynamo_streams().list_streams().get('Streams')

class Dynamo_Table:
    def __init__(self,table_name, key):
        self.table_name = table_name
        self.key        = key
        self.dynamo     = Dynamo()
        self.resource   = self.dynamo.resource
        self.table      = self.resource.Table(self.table_name)
        self.chuck_size = 100

        self._keys      = None

    def add(self,row):
        self.table.put_item(Item=row)
        return row

    def add_batch(self, items):
        with self.table.batch_writer() as batch:
            for item in items:
                batch.put_item(Item=item)

    def delete(self, item):
        self.table.delete_item(Key= { self.key : item })

    def delete_batch(self, items):
        with self.table.batch_writer() as batch:
            for item in items:
                batch.delete_item(Key= item )

    def exists(self):
        if self.info():
            return True
        return False

    def get(self, key_value):
        data = self.table.get_item(Key={self.key: key_value})
        return data.get('Item')

    def get_batch(self, keys):
        def chunks(items,split):
            for i in range(0, len(items), split):
                yield items[i:i + split]

        for chunck in chunks(keys, self.chuck_size):
            request_items = []
            for key in chunck:
                request_items.append({self.key:  key })
            response = self.resource.batch_get_item(RequestItems={self.table_name: {'Keys': request_items}})
            yield response['Responses'][self.table_name]

    def info(self):
        try:
            return self.dynamo.dynamo().describe_table(TableName = self.table_name)
        except:
            return None

    def keys(self, use_cache=True):
        if self._keys and use_cache:
            return self._keys
        def get_scan_batch (start_key):
            if start_key:
                return self.table.scan(ProjectionExpression=self.key, ExclusiveStartKey =start_key );
            else:
                return self.table.scan(ProjectionExpression=self.key)

        keys = []

        def map_Items(items):
            for item in items:
                 keys.append(item[self.key])
            #print(len(items))

        response = get_scan_batch(None)                                 # first call
        map_Items(response['Items'])                                    # map items
        while 'LastEvaluatedKey' in response:                           # see if we need to page
            response = get_scan_batch(response.get('LastEvaluatedKey')) # call with LastEvaluatedKey
            map_Items(response['Items'])                                # map items

        self._keys = keys

        return keys

    def stream_arn(self):
        streams   = self.dynamo.dynamo_streams().list_streams(TableName=self.table_name).get('Streams')
        first_one = Misc.array_pop(streams,0)
        return Misc.get_value(first_one,'StreamArn')

    def stream_info(self):
        stream_arn = self.stream_arn()
        if stream_arn:
            return self.dynamo.dynamo_streams().describe_stream(StreamArn=stream_arn).get('StreamDescription')

    def stream_get_data_latest(self):
        stream_info    = self.stream_info()
        shard_id       = stream_info.get('Shards').pop(0).get('ShardId')
        shard_iterator = self.dynamo.dynamo_streams().get_shard_iterator(StreamArn=self.stream_arn(), ShardId=shard_id, ShardIteratorType='LATEST').get('ShardIterator')
        return           self.dynamo.dynamo_streams().get_records(ShardIterator=shard_iterator)

        #shard_iterator = get_shard_iterator

    def status(self):
        return self.info()['Table']['TableStatus']


import json

import boto3
from pbx_gs_python_utils.utils.Dev import Dev
from pbx_gs_python_utils.utils.Misc import Misc

from osbot_aws.Globals import Globals
from osbot_aws.apis.Session import Session


class IAM:

    def __init__(self, user_name=None, role_name=None):
        self._iam        = None
        self._sts        = None
        self._account_id = None
        self.user_name   = user_name
        self.role_name   = role_name

    # helpers
    def iam(self):
        if self._iam is None:
            self._iam = Session().client('iam')
        return self._iam

    def sts(self):
        if self._sts is None:
            self._sts = Session().client('sts')
        return self._sts

    # main method

    def account_id(self, profile_name=None):
        if profile_name is not None:                            # if profile_name is set
            Globals.aws_session_profile_name = profile_name     # set it globally (since this will be used by all boto3 clients)
        if self._account_id is None:
            self._account_id = self.caller_identity().get('Account')
        return self._account_id

    # def assume_role(self, role_arn,role_session_name):
    #     data = self.sts().assume_role(RoleArn = role_arn, RoleSessionName=role_session_name)
    #     return data

    def caller_identity(self):
        data = self.sts().get_caller_identity()
        del data['ResponseMetadata']
        return data

    def get_data(self, method, field_id, use_paginator, **kwargs):
        paginator = self.iam().get_paginator(method)
        for page in paginator.paginate(**kwargs):
            for id in page.get(field_id):
                yield id
            if use_paginator is False:
                return

    def groups(self):
        return list(self.get_data('list_groups', 'Groups', True))

    def policy_arn(self, policy_name, policy_path='/',account_id=None):
        if policy_name is None: return policy_name
        if account_id is None: account_id = self.account_id()
        if policy_path is None or policy_path == '/':
            return 'arn:aws:iam::{0}:policy/{1}'.format(account_id, policy_name)
        return 'arn:aws:iam::{0}:policy{1}/{2}' .format(account_id, policy_path, policy_name)

    def policy_create_version(self, policy_name, policy_document, policy_path='/', account_id=None): # todo: add support for deleting previous versions
        policy_arn = self.policy_arn(policy_name, policy_path, account_id)
        if type(policy_document) is not str:
            policy_document = json.dumps(policy_document)
        data = self.iam().create_policy_version(PolicyArn=policy_arn, PolicyDocument=policy_document,SetAsDefault=True).get('Policy')
        return {'status': 'ok', 'policy_name': policy_name, 'policy_arn': policy_arn, 'data': data}

    def policy_create(self, policy_name, policy_document, policy_path='/', account_id=None, delete_before_create=False):
        policy_arn = self.policy_arn(policy_name, policy_path,account_id)
        if delete_before_create:
            self.policy_delete(policy_arn)
        if self.policy_exists(policy_arn) is False:
            if type(policy_document) is not str:
                policy_document = json.dumps(policy_document)
            try:
                data       = self.iam().create_policy(PolicyName=policy_name, PolicyDocument=policy_document).get('Policy')
                policy_arn = data.get('Arn')
                return {'status': 'ok'     , 'policy_name': policy_name, 'policy_arn': policy_arn                  , 'data': data                     }
            except Exception as error:
                return {'status': 'error'  , 'policy_name': policy_name, 'policy_arn': None                        , 'data': '{0}'.format(error)      }
        return {        'status': 'warning', 'policy_name': policy_name, 'policy_arn': policy_arn                  , 'data': 'policy already existed' }

    def policy_delete(self, policy_arn):
        if self.policy_exists(policy_arn) is False: return False
        self.policy_detach_roles(policy_arn)
        self.iam().delete_policy(PolicyArn= policy_arn)
        return self.policy_exists(policy_arn) is False

    def policy_delete_by_name(self, policy_name, policy_path='/',account_id=None):
        policy_arn = self.policy_arn(policy_name,policy_path,account_id)
        if self.policy_exists(policy_arn) is False: return False
        self.iam().delete_policy(PolicyArn= policy_arn)
        return self.policy_exists(policy_arn) is False

    def policy_info(self, policy_arn):
        try:
            return self.iam().get_policy(PolicyArn=policy_arn).get('Policy')
        except:
            return None

    def policy_details(self, policy_arn):
        try:
            policy_info    = self.policy_info(policy_arn)
            if policy_info:
                policy_arn     = policy_info.get('Arn')
                version_id     = policy_info.get('DefaultVersionId')
                policy_version = self.iam().get_policy_version(PolicyArn=policy_arn, VersionId=version_id).get('PolicyVersion')
                return {'policy_arn': policy_arn, 'policy_name': policy_info.get('PolicyName'), 'policy_info' : policy_info, 'policy_version': policy_version}
        except:
            return None

    def policy_detach_roles(self,policy_arn):
        policy_roles = self.iam().list_entities_for_policy(PolicyArn=policy_arn).get('PolicyRoles')
        for role in policy_roles:
            IAM(role_name=role.get('RoleName')).role_policy_detach(policy_arn)
        return self

    def policy_statement(self, policy_arn):
        policy_details = self.policy_details(policy_arn)
        if policy_details:
            return policy_details.get('policy_version').get('Document').get('Statement')

    def policy_exists(self, policy_arn):
        return self.policy_info(policy_arn) is not None

    def policy_exists_by_name(self, policy_name, policy_path='/', account_id=None):
        calculated_policy_arn = self.policy_arn(policy_name, policy_path, account_id)   # since there doesn't seem to be a way to search for a policy by name
        return self.policy_info(calculated_policy_arn) is not None

    def policies(self):
        return list(self.get_data('list_policies', 'Policies', True))

    def policies_delete(self, policies_names):
        for policy_name in policies_names:
            self.policy_delete(policy_name)

    def role_arn(self):
        return Misc.get_value(self.role_info(), 'Arn')

    def role_exists(self):
        return self.role_info() is not None

    def role_info(self):
        try:
            return self.iam().get_role(RoleName=self.role_name).get('Role')
        except:
            return None

    def role_create(self, policy_document):
        if type(policy_document) is not str:
            policy_document = json.dumps(policy_document)
        if self.role_exists() is False:
            return self.iam().create_role(RoleName=self.role_name, AssumeRolePolicyDocument=policy_document).get('Role')

    def role_delete(self):
        if self.role_exists() is False: return False                # make sure it exists
        self.role_policies_detach_and_delete()                      # delete all associated policies
        self.iam().delete_role(RoleName=self.role_name)
        return self.role_exists() is False

    def role_policies_detach_and_delete(self):
        policies_arns = self.role_policies().values()
        self.role_policies_detach(policies_arns)
        self.policies_delete     (policies_arns)

    def role_policy_attach(self,policy_arn):
        return self.iam().attach_role_policy(RoleName=self.role_name, PolicyArn=policy_arn)

    def role_policy_detach(self, policy_arn):
        self.iam().detach_role_policy(RoleName=self.role_name, PolicyArn=policy_arn)

    def role_policies_detach(self, policies_arn):
        for policy_arn in policies_arn:
            self.role_policy_detach(policy_arn)


    def role_policies_attach(self, policies_arn):
        for policy_arn in policies_arn:
            self.role_policy_attach(policy_arn)

    def role_policies(self):
        policies = {}
        if self.role_name:
            for item in self.get_data('list_attached_role_policies', 'AttachedPolicies', True, RoleName=self.role_name):
                policies[item.get('PolicyName')] = item.get('PolicyArn')
        return policies

    def role_policies_statements(self, just_statements = False):
        if just_statements:
            statements = []
            for items in self.role_policies_statements().values():
                statements.extend(items)
            return statements

        policy_statements = {}
        for policy_name, policy_arn in self.role_policies().items():
            policy_statements[policy_name] = self.policy_statement(policy_arn)
        return policy_statements


    def roles(self):
        return list(self.get_data('list_roles', 'Roles', True))

    def users(self):
        return list(self.get_data('list_users', 'Users', True))

    def user_exists(self):
        return self.user_info() is not None

    def user_info(self):
        try:
            return self.iam().get_user(UserName=self.user_name).get('User')
        except:
            return None

    def user_create(self):
        if self.user_exists() is False:
            return self.iam().create_user(UserName=self.user_name).get('User')

    def user_delete(self):
        if self.user_exists() is False: return False
        self.iam().delete_user(UserName=self.user_name)
        return self.user_exists() is False

    def set_account_id(self, value): self._account_id = value ;return self
    def set_user_name (self, value): self.user_name   = value ;return self
    def set_role_name (self, value): self.role_name   = value; return self



import json

from pbx_gs_python_utils.utils.Misc import Misc

from osbot_aws.apis.Session import Session


class Queue:
    def __init__(self, queue_name=None, url=None):
        self._url       = url
        self._sqs       = None
        self.queue_name = queue_name

    # helper methods
    def sqs(self):
        if self._sqs is None:
            self._sqs = Session().client('sqs')
        return self._sqs

    def url(self):
        if self._url is None:
            self._url =  self.sqs().get_queue_url(QueueName=self.queue_name).get('QueueUrl')
        return self._url


    # main methods

    def add(self, body, attributes=None):
        self.message_send(body,attributes)
        return self

    def arn(self):
        return self.attributes().get('QueueArn')

    def attributes(self):
        try:
            return self.sqs().get_queue_attributes(QueueUrl=self.url(),AttributeNames=['All']).get('Attributes')
        except:
            return None

    def attributes_update(self, new_attributes):
        return self.sqs().set_queue_attributes(QueueUrl=self.url(), Attributes=new_attributes)


    def create_raw(self,attributes=None):
        if attributes is None: attributes = {}
        return self.sqs().create_queue(QueueName=self.queue_name,Attributes=attributes).get('QueueUrl')

    def create(self,attributes=None):
        queue_url = self.create_raw(attributes)
        if queue_url:
            return self
        return None

    def delete(self):
        if self.exists() is False: return False
        self.sqs().delete_queue(QueueUrl=self.url())
        return self.exists() is False

    def exists(self):
        return self.attributes() is not None

    def get_message(self, delete_message=True):
        message = self.message_raw()
        if message:
            body = message.get('Body')
            receipt_handle = message.get('ReceiptHandle')
            if delete_message:
                self.message_delete(receipt_handle)
            return body

    def get_message_with_attributes(self, delete_message=True):
        message = self.message_raw()
        if message:
            body            = message.get('Body')
            attributes      = message.get('MessageAttributes')
            receipt_handle  = message.get('ReceiptHandle')
            attributes_data = {}
            if attributes:
                for key,value in attributes.items():
                    attributes_data[key] = value.get('StringValue')
            if delete_message:
                self.message_delete(receipt_handle)
            return body, attributes_data

    def get_n_message(self, n, delete_message=True):
        messages = []
        for i in range(0,n):
            message = self.get_message(delete_message)
            if message:
                messages.append(message)
            else:
                break
        return messages

    def list(self):
        return self.sqs().list_queues().get('QueueUrls')


    def message_raw(self):
        messages = self.sqs().receive_message(QueueUrl=self.url(), MessageAttributeNames=['All']).get('Messages')
        return Misc.array_pop(messages,0)

    def message_delete(self, receipt_handle):
        return self.sqs().delete_message(QueueUrl=self.url(), ReceiptHandle=receipt_handle)


    #def add(self, message, attributes=None): return self.message_send(message,attributes)

    def message_send(self,body,attributes_data=None):
        if attributes_data is None:
            return self.sqs().send_message(QueueUrl=self.url(), MessageBody=body).get('MessageId')
        else:
            attributes = {}
            for key,value in attributes_data.items():
                attributes[key] = { 'StringValue': value , 'DataType': 'String'}
            return self.sqs().send_message(QueueUrl=self.url(),MessageBody=body, MessageAttributes=attributes).get('MessageId')

    def messages_in_queue(self):
        return int(self.attributes().get('ApproximateNumberOfMessages'))

    def messages_not_visible(self):
        return int(self.attributes().get('ApproximateNumberOfMessagesNotVisible'))

    def push(self, data):
        if data:
            self.message_send(json.dumps(data))
        return self

    def pull(self, delete_message=True):
        data_json = self.get_message(delete_message=delete_message)
        if data_json:
            return json.loads(data_json)
        return None

    def set_queue_name(self, queue_name):
        self.queue_name = queue_name
        self._url       = None          # need to reset this value or the old value will still be used
        return self


from osbot_aws.apis.Queue import Queue


def run(event, context):
    queue_url = event.get('queue_url')
    message   = event.get('message'  )
    Queue(url=queue_url).push(message)
    return {'status': 'ok'}


import boto3
from botocore.session import get_session

from osbot_aws.Globals import Globals


class Session:

    def client_boto3(self,service_name):
        try:
            profile_name = Globals.aws_session_profile_name
            region_name  = Globals.aws_session_region_name

            profiles = get_session()._build_profile_map()
            if profile_name in profiles:
                session = boto3.Session(profile_name=profile_name, region_name=region_name)
                return {'status': 'ok', 'client': session.client(service_name=service_name) }
            return { 'status' : 'ok', 'client': boto3.client(service_name=service_name)}
        except Exception as error:
            return {'status': 'error', 'data': '{0}'.format(error) }


    def resource_boto3(self,service_name):          # refactor with client_boto3
        try:
            profile_name = Globals.aws_session_profile_name
            region_name  = Globals.aws_session_region_name

            profiles = get_session()._build_profile_map()
            if profile_name in profiles:
                session = boto3.Session(profile_name=profile_name, region_name=region_name)
                return {'status': 'ok', 'resource': session.resource(service_name=service_name) }
            return { 'status' : 'ok', 'resource': boto3.resource(service_name=service_name)}
        except Exception as error:
            return {'status': 'error', 'data': '{0}'.format(error) }

    def client(self, service_name):
        return self.client_boto3(service_name).get('client')

    def resource(self, service_name):
        return self.resource_boto3(service_name).get('resource')



import subprocess
from pbx_gs_python_utils.utils.Process import Process


def run(event, context):
    port  = event.get('port')
    count = event.get('count')
    if port and count:
        for i in range(0,count):
            subprocess.Popen(["python", "-m", "http.server", str(port+i),"&"], cwd='/tmp')

    return Process.run('ps',['-aux']).get('stdout')



from osbot_aws.apis.Queue import Queue


def run(event, context):
    queue_2 = Queue(url='https://sqs.eu-west-2.amazonaws.com/244560807427/unit_tests_temp_queue')
    queue_2.push(event)
    return {'status': 'ok'}


from pbx_gs_python_utils.utils.Misc import Misc


class Temp_Misc:

    @staticmethod
    def get_field(target, field, default=None):
        if target is not None:
            try:
                value = getattr(target, field)
                if value is not None:
                    return value
            except:
                pass
        return default

    @staticmethod
    def get_missing_fields(target,field):
        missing_fields = []
        for field in field:
            if Temp_Misc.get_field(target, field) is None:
                missing_fields.append(field)
        return missing_fields

    @staticmethod
    def last_letter(text):
        if text and (type(text) is str) and len(text) > 0:
            return text[-1]

    @staticmethod
    def random_text(prefix=None,length=12):
        if prefix is None: prefix = 'text_'
        if Temp_Misc.last_letter(prefix) != '_':
            prefix += '_'
        return Misc.random_string_and_numbers(length=length, prefix=prefix)



class Aws_Utils:

    @staticmethod
    def run_code_in_lambda(code):
        file_Path = 'temp_code/code.py'
        temp_Dir  = 'temp_code'
        zip_file  = 'temp_code.zip'

        def create_temp_files():
            if not os.path.exists(temp_Dir):
                os.mkdir(temp_Dir)
            with open(file_Path, "w+") as f:
                f.write(code)

        def delete_temp_files():
            os.remove(file_Path)
            os.remove(zip_file)
            os.rmdir(temp_Dir)

        create_temp_files()

        name      = 'dynamic_code'
        role      = 'arn:aws:iam::244560807427:role/lambda_basic_execution'
        handler   = 'code.dynamic_method'
        s3_bucket = 'gs-lambda-tests'
        s3_key    = 'dinis/lambda-using-dynamic-code.zip'

        aws = Aws_Cli()
        aws.lambda_delete_function(name)
        aws.s3_upload_folder      (temp_Dir, s3_bucket, s3_key)
        aws.lambda_create_function(name, role, handler, s3_bucket, s3_key)

        (result, response) = aws.lambda_invoke_function(name, {})

        aws.lambda_delete_function(name)
        delete_temp_files()
        return result

    @staticmethod
    def zip_folder(root_dir):
        return shutil.make_archive(root_dir, "zip", root_dir)